#include <winsock2.h>
#include <windows.h>
#include <winreg.h>
#include <commctrl.h>
#include "resource.h"

/* various constants */
#define DEFAULT_HOST	""
#define DEFAULT_PORT	80
#define DEFAULT_FLASH	300
#define DEFAULT_POLL	60
#define DEFAULT_RESOLVE	TRUE

#define STATUS_TRYING	0
#define STATUS_OFFLINE	1
#define STATUS_ONLINE	2
#define STATUS_ERROR	3
#define N_STATUSES		4

#define N_ICONS			3

#define TIMER_POLL		1
#define TIMER_FLASH		2

#define UII_YELLOW		0
#define UII_GREEN		1
#define UII_RED			2
#define UIF_FLASH		0x80

/* global variables */
struct
{
	/* remote host */
	char			*pHost;
	unsigned short	nPort;

	/* intervals */
	unsigned long	uPoll;
	unsigned long	uFlash;

	/* behaviors */
	unsigned char	aBehaviors[N_STATUSES];

	/* misc */
	BOOL			bResolve;
} g_oOptions;
HINSTANCE		g_hInst;
HWND			g_hWnd;
HMENU			g_hMenu;
HMENU			g_hPopupMenu;
NOTIFYICONDATA	g_NID;
HIMAGELIST		g_hImageList;
unsigned int	g_nStatus;
SOCKET			g_s;
struct sockaddr_in g_sin;
HICON			g_hOldIcon;
HANDLE			g_hLookup;
char			g_aLookupBuf[MAXGETHOSTSTRUCT];
struct
{
	unsigned short usPort;
	const char *pszName;
} g_aDefaultPorts[] = {
	{21,	"21 (FTP)"},
	{22,	"22 (SSH)"},
	{23,	"23 (Telnet)"},
	{25,	"25 (SMTP)"},
	{80,	"80 (HTTP)"},
	{110,	"110 (POP3)"},
	{0,		NULL}
};
const unsigned char	g_aDefaultBehaviors[N_STATUSES] = {
	/* TRYING */	UII_YELLOW,
	/* OFFLINE */	UII_RED,
	/* ONLINE */	UII_GREEN | UIF_FLASH,
	/* ERROR */		UII_RED | UIF_FLASH,
};
const unsigned long g_aIcons[N_ICONS] = {IDI_YELLOW, IDI_GREEN, IDI_RED};
char * const		g_aIconNames[N_ICONS] = {"Yellow", "Green", "Red"};

/* function declarations */
BOOL CALLBACK ConfigProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void LoadOptions();
void SaveOptions();
void ShowOptions();
BOOL GetOptions();
void SetStatus(unsigned int nStatus, char *pStatus, BOOL bReFlash);

BOOL LoadStringOption(HKEY hKey, const char *pszSubKey, char **ppBuf);
BOOL LoadShortOption(HKEY hKey, const char *pszSubKey, unsigned short *pShort);
BOOL LoadLongOption(HKEY hKey, const char *pszSubKey, unsigned long *pLong);
BOOL LoadBoolOption(HKEY hKey, const char *pszSubKey, BOOL *pBool);

/* macros */
/*int g_aStatusIcon[] = {IDI_YELLOW, IDI_RED, IDI_GREEN, IDI_RED};*/
#define STATUSICONINDEX(s)	(g_oOptions.aBehaviors[(s)] & ~UIF_FLASH)
#define STATUSICON(s)		(g_aIcons[STATUSICONINDEX((s))])
#define STATUSFLASH(s)		(g_oOptions.aBehaviors[(s)] & UIF_FLASH)
#define SWAP(t,a,b)			do {t __temp; __temp = (a); (a) = (b); (b) = __temp;} while(0)

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nShowCmd)
{
	INITCOMMONCONTROLSEX iccx;
	WSADATA wsaData;
	int iRet;

	g_hInst = hInst;

	iccx.dwSize = sizeof(iccx);
	iccx.dwICC	= ICC_USEREX_CLASSES;
	InitCommonControlsEx(&iccx);

	if(WSAStartup(MAKEWORD(1, 1), &wsaData) == 0)
	{
		iRet = DialogBox(hInst, MAKEINTRESOURCE(IDD_CONFIG), NULL, ConfigProc);
		WSACleanup();
		return iRet;
	}
	else
	{
		MessageBox(NULL, "Error starting Winsock!", NULL, MB_ICONSTOP);
		return 1;
	}
}

BOOL CALLBACK ConfigProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg)
	{
	case WM_INITDIALOG:
		/* initialize variables */
		g_nStatus			= STATUS_TRYING;
		g_hWnd				= hWnd;
		g_hOldIcon			= NULL;
		g_hLookup			= NULL;
		g_hMenu				= LoadMenu(g_hInst, MAKEINTRESOURCE(IDR_POPUP));
		g_hPopupMenu		= GetSubMenu(g_hMenu, 0);
		g_hImageList		= ImageList_Create(16, 16, ILC_COLOR4 | ILC_MASK, 3, 0);
		SetMenuDefaultItem(g_hPopupMenu, IDM_CONFIG, FALSE);
		g_s					= INVALID_SOCKET;
		g_sin.sin_family	= AF_INET;
		g_sin.sin_addr.S_un.S_addr = INADDR_NONE;
		ZeroMemory(g_sin.sin_zero, sizeof(g_sin.sin_zero));

		/* fill imagelist */
		ImageList_AddIcon(g_hImageList, LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_YELLOW)));
		ImageList_AddIcon(g_hImageList, LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_GREEN)));
		ImageList_AddIcon(g_hImageList, LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_RED)));

		/* load & show options */
		{
			int i;

			for(i = 0; g_aDefaultPorts[i].usPort; i++)
			{
				SendDlgItemMessage(hWnd, IDC_PORT, CB_SETITEMDATA,
					SendDlgItemMessage(hWnd, IDC_PORT, CB_ADDSTRING, 0,
						(LPARAM)(g_aDefaultPorts[i].pszName)),
					g_aDefaultPorts[i].usPort);
			}

			/* add items to comboboxex */
			SendDlgItemMessage(hWnd, IDC_TRYING, CBEM_SETIMAGELIST, 0, (LPARAM)g_hImageList);
			SendDlgItemMessage(hWnd, IDC_OFFLINE, CBEM_SETIMAGELIST, 0, (LPARAM)g_hImageList);
			SendDlgItemMessage(hWnd, IDC_ONLINE, CBEM_SETIMAGELIST, 0, (LPARAM)g_hImageList);
			SendDlgItemMessage(hWnd, IDC_ERROR, CBEM_SETIMAGELIST, 0, (LPARAM)g_hImageList);

			for(i = 0; i < N_ICONS; i++)
			{
				COMBOBOXEXITEM cbxi;

				cbxi.mask		= CBEIF_IMAGE | CBEIF_SELECTEDIMAGE | CBEIF_TEXT;
				cbxi.iItem		= i;
				cbxi.pszText	= g_aIconNames[i];
				cbxi.iImage		= i;
				cbxi.iSelectedImage = i;

				SendDlgItemMessage(hWnd, IDC_TRYING, CBEM_INSERTITEM, 0, (LPARAM)&cbxi);
				SendDlgItemMessage(hWnd, IDC_OFFLINE, CBEM_INSERTITEM, 0, (LPARAM)&cbxi);
				SendDlgItemMessage(hWnd, IDC_ONLINE, CBEM_INSERTITEM, 0, (LPARAM)&cbxi);
				SendDlgItemMessage(hWnd, IDC_ERROR, CBEM_INSERTITEM, 0, (LPARAM)&cbxi);
			}
		}
		
		LoadOptions();
		ShowOptions();
		EnableWindow(GetDlgItem(hWnd, IDAPPLY), FALSE);

		/* add tray icon */
		g_NID.cbSize		= sizeof(g_NID);
		g_NID.hWnd			= hWnd;
		g_NID.uID			= 0;
		g_NID.uFlags		= NIF_ICON | NIF_MESSAGE | NIF_TIP;
		g_NID.uCallbackMessage = WM_USER;
		g_NID.hIcon			= LoadIcon(g_hInst, MAKEINTRESOURCE(STATUSICON(g_nStatus)));
		strncpy(g_NID.szTip, "Initializing...", sizeof(g_NID.szTip));
		if(!Shell_NotifyIcon(NIM_ADD, &g_NID))
		{
			MessageBox(hWnd, "Error adding tray icon!", NULL, MB_ICONSTOP);
			EndDialog(hWnd, 1);
		}

		/* behaviors take effect immediately: start flashing if necessary */
		SetStatus(g_nStatus, NULL, TRUE);

		break;
	case WM_TIMER:
		switch(wParam)
		{
		case TIMER_POLL:
			/* tell the user what we're up to */
			SetStatus(STATUS_TRYING, "Trying to connect...", FALSE);

			/* close any stale connections */
			if(g_s != INVALID_SOCKET)
				closesocket(g_s);

			g_s = socket(AF_INET, SOCK_STREAM, 0);
			if(g_s == INVALID_SOCKET)
			{
				SetStatus(STATUS_ERROR, "Error creating socket!", FALSE);
				break;
			}

			/* fill in sockaddr_in */
			g_sin.sin_port = htons(g_oOptions.nPort);

			/* only recompute IP address if we don't already have it */
			if(g_sin.sin_addr.S_un.S_addr == INADDR_NONE)
			{
				/* see if the host strings is an IP address */
				if((g_sin.sin_addr.S_un.S_addr = inet_addr(g_oOptions.pHost)) == INADDR_NONE)
				{
					g_hLookup = WSAAsyncGetHostByName(hWnd, WM_USER, g_oOptions.pHost,
						g_aLookupBuf, sizeof(g_aLookupBuf));
					if(!g_hLookup)
					{
						SetStatus(STATUS_ERROR, "WSAAsyncGetHostByName() failed!", FALSE);
					}
					break;
				}
			}

Connect:
			/* set up asynchronous notification */
			if(WSAAsyncSelect(g_s, hWnd, WM_USER, FD_CONNECT) != 0)
			{
				SetStatus(STATUS_ERROR, "WSAAsyncSelect() failed!", FALSE);
				break;
			}

			/* kick off a connection attempt */
			if(connect(g_s, (struct sockaddr *)&g_sin, sizeof(g_sin)) == SOCKET_ERROR
				&& WSAGetLastError() != WSAEWOULDBLOCK)
			{
				SetStatus(STATUS_ERROR, "connect() failed!", FALSE);
				break;
			}

			/* just wait for some event to pop up in WM_USER */

			break;
		case TIMER_FLASH:
			if(STATUSFLASH(g_nStatus))
			{
				SWAP(HICON, g_hOldIcon, g_NID.hIcon);
				Shell_NotifyIcon(NIM_MODIFY, &g_NID);
			}
			break;
		}
		break;
	case WM_COMMAND:
		switch(HIWORD(wParam))
		{
		case CBN_EDITCHANGE:
		case CBN_SELCHANGE:
		case EN_CHANGE:
			EnableWindow(GetDlgItem(hWnd, IDAPPLY), TRUE);
			break;
		case BN_CLICKED:
			switch(LOWORD(wParam))
			{
			case IDOK:
				/* simulate clicking Apply button */
				SendMessage(hWnd, WM_COMMAND, MAKEWORD(IDAPPLY, BN_CLICKED), (LPARAM)GetDlgItem(hWnd, IDAPPLY));

				/* start polling right when the user hits OK:
				   set the timer, and kick off a connection right away
				   (the timer proc closes the socket automatically) */
				SetTimer(hWnd, TIMER_POLL, g_oOptions.uPoll * 1000, NULL);
				PostMessage(hWnd, WM_TIMER, TIMER_POLL, 0);

				/* pass-through and hide the window */
			case IDCANCEL:
				/* hide window */
				ShowWindow(hWnd, SW_HIDE);
				break;
			case IDAPPLY:
				/* save options and disable button */
				if(!GetOptions())
				{
					MessageBeep(0xffffffff);
					break;
				}
				SaveOptions();

				EnableWindow(GetDlgItem(hWnd, IDAPPLY), FALSE);
				break;
			case IDEXIT:
				/* simulate clicking close button */
				SendMessage(hWnd, WM_CLOSE, 0, 0);
				break;
			default:
				/* chances are a checkbox was clicked */
				EnableWindow(GetDlgItem(hWnd, IDAPPLY), TRUE);
			}
			break;
		}
		break;
	case WM_USER:
		/* we have to be SURE if the message came from Winsock or not, because
		   WM_USER is also used for tray icon notifications */
		if(g_s != INVALID_SOCKET && wParam == (WPARAM)g_s && WSAGETSELECTEVENT(lParam) == FD_CONNECT)
		{
			switch(WSAGETSELECTERROR(lParam))
			{
			case 0:
				/* woohoo, host's up! */
				SetStatus(STATUS_ONLINE, "The connection has been accepted.", FALSE);
				break;
			case WSAECONNREFUSED:
				/* "The attempt to connect was forcefully rejected." */
				SetStatus(STATUS_OFFLINE, "The connection attempt was rejected.", FALSE);
				break;
			case WSAENETUNREACH:
				/* "The network cannot be reached from this host at this time." */
				SetStatus(STATUS_OFFLINE, "The network is unreachable at this time.", FALSE);
				break;
			case WSAETIMEDOUT:
				/* "Attempt to connect timed out without establishing a connection." */
				SetStatus(STATUS_OFFLINE, "The connection timed out.", FALSE);
				break;
			default:
				SetStatus(STATUS_ERROR, "An unexpected Winsock error occured!", FALSE);
				break;
			}

			/* regardless of the outcome, clean up */
			shutdown(g_s, SD_BOTH);
			closesocket(g_s);

			/* if we want to resolve the next time we connect, clear the address */
			if(g_oOptions.bResolve)
				g_sin.sin_addr.S_un.S_addr = INADDR_NONE;
		}
		else if(g_hLookup != NULL && wParam == (WPARAM)g_hLookup)
		{
			switch(WSAGETASYNCERROR(lParam))
			{
			case 0:
				/* the host was found, get IP address information and try to connect */
				g_sin.sin_addr.S_un.S_addr = *((unsigned long *)
					((((struct hostent *)g_aLookupBuf)->h_addr_list)[0]));

				/* YES, A GOTO! */
				goto Connect;
			case WSAHOST_NOT_FOUND:
				/* "Authoritative Answer Host not found." */
				SetStatus(STATUS_OFFLINE, "The host was not found by the DNS server.", FALSE);
				break;
			case WSATRY_AGAIN:
				/* "Non-Authoritative Host not found, or SERVERFAIL." */
				SetStatus(STATUS_OFFLINE, "A recoverable error occured, trying again later.", FALSE);
				break;
			case WSANO_RECOVERY:
				/* "Nonrecoverable errors, FORMERR, REFUSED, NOTIMP." */
				SetStatus(STATUS_OFFLINE, "A non-recoverable error occured, trying again later anyway.", FALSE);
				break;
			case WSANO_DATA:
				/* "Valid name, no data record of requested type." */
				SetStatus(STATUS_OFFLINE, "The host doesn't appear to have an IPv4 DNS record.", FALSE);
				break;
			default:
				SetStatus(STATUS_ERROR, "An unexpected Winsock/DNS error occured!", FALSE);
				break;
			}
			
			g_hLookup = NULL;
		}
		else if(wParam == g_NID.uID)
		{
			POINT pt;

			switch(lParam)
			{
			case WM_LBUTTONDBLCLK:
				/* make sure the saved options are up there */
				ShowOptions();
				EnableWindow(GetDlgItem(hWnd, IDAPPLY), FALSE);

				ShowWindow(hWnd, SW_SHOW);
				ShowWindow(hWnd, SW_SHOWNORMAL);
				SetActiveWindow(hWnd);
				break;
			case WM_RBUTTONUP:
				/* display a nice popup menu */
				GetCursorPos(&pt);
				SetActiveWindow(hWnd);
				switch(TrackPopupMenuEx(g_hPopupMenu, TPM_RIGHTALIGN | TPM_BOTTOMALIGN |
					TPM_NONOTIFY | TPM_RETURNCMD | TPM_RIGHTBUTTON, pt.x, pt.y, hWnd, NULL))
				{
				case IDM_CONFIG:
					/* simulate left double-click */
					SendMessage(hWnd, WM_USER, g_NID.uID, WM_LBUTTONDBLCLK);
					break;
				case IDM_EXIT:
					/* simulate clicking the close button */
					SendMessage(hWnd, WM_CLOSE, 0, 0);
					break;
				}
				break;
			}
		}
		break;
	case WM_SIZE:
		if(wParam == SIZE_MINIMIZED)
		{
			/* minimize -> hide window */
			ShowWindow(hWnd, SW_HIDE);
		}
		break;
	case WM_CLOSE:
		if(MessageBox(hWnd, "Are you sure you want to exit?", "Question",
			MB_YESNO | MB_ICONQUESTION) == IDYES)
		{
			/* store options in registry */
			SaveOptions();

			/* do cleanup work */
			if(g_s != INVALID_SOCKET)
				closesocket(g_s);

			if(g_hLookup)
				WSACancelAsyncRequest(g_hLookup);
			
			free(g_oOptions.pHost);

			DestroyMenu(g_hMenu);

			KillTimer(hWnd, TIMER_POLL);
			if(STATUSFLASH(g_nStatus))
				KillTimer(hWnd, TIMER_FLASH);

			Shell_NotifyIcon(NIM_DELETE, &g_NID);
			EndDialog(hWnd, 0);
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

/* displays options in the dialog */
void ShowOptions()
{
	char szPort[8];

	/* host informations */
	SetDlgItemText(g_hWnd, IDC_HOST, g_oOptions.pHost);

	wsprintf(szPort, "%d", g_oOptions.nPort);
	if(SendDlgItemMessage(g_hWnd, IDC_PORT, CB_SELECTSTRING, -1, (LPARAM)szPort) == CB_ERR)
	{
		SetDlgItemText(g_hWnd, IDC_PORT, szPort);
	}

	/* intervals */
	SetDlgItemInt(g_hWnd, IDC_POLL, g_oOptions.uPoll, FALSE);
	SetDlgItemInt(g_hWnd, IDC_FLASH, g_oOptions.uFlash, FALSE);

	/* behaviors */
	SendDlgItemMessage(g_hWnd, IDC_TRYING, CB_SETCURSEL, STATUSICONINDEX(STATUS_TRYING), 0);
	SendDlgItemMessage(g_hWnd, IDC_OFFLINE, CB_SETCURSEL, STATUSICONINDEX(STATUS_OFFLINE), 0);
	SendDlgItemMessage(g_hWnd, IDC_ONLINE, CB_SETCURSEL, STATUSICONINDEX(STATUS_ONLINE), 0);
	SendDlgItemMessage(g_hWnd, IDC_ERROR, CB_SETCURSEL, STATUSICONINDEX(STATUS_ERROR), 0);
	CheckDlgButton(g_hWnd, IDC_TRYINGFLASH, STATUSFLASH(STATUS_TRYING) ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(g_hWnd, IDC_OFFLINEFLASH, STATUSFLASH(STATUS_OFFLINE) ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(g_hWnd, IDC_ONLINEFLASH, STATUSFLASH(STATUS_ONLINE) ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(g_hWnd, IDC_ERRORFLASH, STATUSFLASH(STATUS_ERROR) ? BST_CHECKED : BST_UNCHECKED);

	/* misc */
	CheckDlgButton(g_hWnd, IDC_RESOLVE, g_oOptions.bResolve ? BST_CHECKED : BST_UNCHECKED);
}

/* retrieves options from the dialog */
BOOL GetOptions()
{
	int nLen, n;

	/* host information */
	nLen = GetWindowTextLength(GetDlgItem(g_hWnd, IDC_HOST)) + 1;
	if(nLen == 1)
	{
		SetFocus(GetDlgItem(g_hWnd, IDC_HOST));
		return FALSE;
	}

	free(g_oOptions.pHost);
	g_oOptions.pHost = malloc(nLen);
	GetDlgItemText(g_hWnd, IDC_HOST, g_oOptions.pHost, nLen);

	n = GetDlgItemInt(g_hWnd, IDC_PORT, NULL, FALSE);
	if(n >> 16 || n == 0) /* if n >= 0x10000 */
	{
		SetFocus(GetDlgItem(g_hWnd, IDC_PORT));
		return FALSE;
	}
	g_oOptions.nPort = (unsigned short)n;

	/* intervals */
	if(!(g_oOptions.uPoll = GetDlgItemInt(g_hWnd, IDC_POLL, NULL, FALSE)))
	{
		SetFocus(GetDlgItem(g_hWnd, IDC_POLL));
		return FALSE;
	}

	if(!(g_oOptions.uFlash = GetDlgItemInt(g_hWnd, IDC_FLASH, NULL, FALSE)))
	{
		SetFocus(GetDlgItem(g_hWnd, IDC_FLASH));
		return FALSE;
	}

	/* if necessary, change flashing interval immediately */
	if(STATUSFLASH(g_nStatus))
	{
		KillTimer(g_hWnd, TIMER_FLASH);
		SetTimer(g_hWnd, TIMER_FLASH, g_oOptions.uFlash, NULL);
		PostMessage(g_hWnd, WM_TIMER, TIMER_FLASH, 0);
	}

	/* behaviors */
	g_oOptions.aBehaviors[STATUS_TRYING] =
		SendDlgItemMessage(g_hWnd, IDC_TRYING, CB_GETCURSEL, 0, 0) |
		((IsDlgButtonChecked(g_hWnd, IDC_TRYINGFLASH)) ? UIF_FLASH : 0);
	g_oOptions.aBehaviors[STATUS_OFFLINE] =
		SendDlgItemMessage(g_hWnd, IDC_OFFLINE, CB_GETCURSEL, 0, 0) |
		((IsDlgButtonChecked(g_hWnd, IDC_OFFLINEFLASH)) ? UIF_FLASH : 0);
	g_oOptions.aBehaviors[STATUS_ONLINE] =
		SendDlgItemMessage(g_hWnd, IDC_ONLINE, CB_GETCURSEL, 0, 0) |
		((IsDlgButtonChecked(g_hWnd, IDC_ONLINEFLASH)) ? UIF_FLASH : 0);
	g_oOptions.aBehaviors[STATUS_ERROR] =
		SendDlgItemMessage(g_hWnd, IDC_ERROR, CB_GETCURSEL, 0, 0) |
		((IsDlgButtonChecked(g_hWnd, IDC_ERRORFLASH)) ? UIF_FLASH : 0);

	/* behavior changes take effect immediately */
	SetStatus(g_nStatus, NULL, TRUE);

	/* to resolve or not to resolve, that is the question */
	g_oOptions.bResolve = (IsDlgButtonChecked(g_hWnd, IDC_RESOLVE) == BST_CHECKED);

	return TRUE;
}

/* loads options from registry */
void LoadOptions()
{
	HKEY hKey;

	if(RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\UpAlert", 0, KEY_QUERY_VALUE, &hKey) != ERROR_SUCCESS)
	{
		g_oOptions.pHost	= strdup(DEFAULT_HOST);
		g_oOptions.nPort	= DEFAULT_PORT;
		g_oOptions.uPoll	= DEFAULT_POLL;
		g_oOptions.uFlash	= DEFAULT_FLASH;
		CopyMemory(g_oOptions.aBehaviors, g_aDefaultBehaviors, sizeof(g_oOptions.aBehaviors));
		g_oOptions.bResolve	= DEFAULT_RESOLVE;
		return;
	}

	if(!LoadStringOption(hKey, "Host", &g_oOptions.pHost))
		g_oOptions.pHost = strdup(DEFAULT_HOST);

	if(!LoadShortOption(hKey, "Port", &g_oOptions.nPort))
		g_oOptions.nPort = DEFAULT_PORT;

	if(!LoadLongOption(hKey, "Poll", &g_oOptions.uPoll))
		g_oOptions.uPoll = DEFAULT_POLL;

	if(!LoadLongOption(hKey, "Flash", &g_oOptions.uFlash))
		g_oOptions.uFlash = DEFAULT_FLASH;

	if(!LoadLongOption(hKey, "Behaviors", (unsigned long *)g_oOptions.aBehaviors))
		CopyMemory(g_oOptions.aBehaviors, g_aDefaultBehaviors, sizeof(g_oOptions.aBehaviors));

	if(!LoadBoolOption(hKey, "Resolve", &g_oOptions.bResolve))
		g_oOptions.bResolve = DEFAULT_RESOLVE;

	RegCloseKey(hKey);
	return;
}

/* saves options to the registry */
void SaveOptions()
{
	DWORD dw;
	HKEY hKey;

	if(RegCreateKeyEx(HKEY_CURRENT_USER, "Software\\UpAlert", 0, NULL,
		REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hKey, &dw) == ERROR_SUCCESS)
	{
		/* try to save as many options possible */
		RegSetValueEx(hKey, "Host", 0, REG_SZ, (LPBYTE)g_oOptions.pHost, strlen(g_oOptions.pHost) + 1);
		
		dw = g_oOptions.nPort;
		RegSetValueEx(hKey, "Port", 0, REG_DWORD, (LPBYTE)&dw, sizeof(dw));

		RegSetValueEx(hKey, "Poll", 0, REG_DWORD, (LPBYTE)&g_oOptions.uPoll, sizeof(g_oOptions.uPoll));

		RegSetValueEx(hKey, "Flash", 0, REG_DWORD, (LPBYTE)&g_oOptions.uFlash, sizeof(g_oOptions.uFlash));

		RegSetValueEx(hKey, "Behaviors", 0, REG_DWORD, (LPBYTE)g_oOptions.aBehaviors, sizeof(g_oOptions.aBehaviors));

		dw = g_oOptions.bResolve;
		RegSetValueEx(hKey, "Resolve", 0, REG_DWORD, (LPBYTE)&dw, sizeof(dw));

		RegCloseKey(hKey);
	}
}

BOOL LoadStringOption(HKEY hKey, const char *pszSubKey, char **ppBuf)
{
	DWORD dwLen, dwType;

	if(RegQueryValueEx(hKey, pszSubKey, NULL, &dwType, NULL, &dwLen) != ERROR_SUCCESS)
		return FALSE;

	if(dwType != REG_SZ)
		return FALSE;

	*ppBuf = (char *)malloc(dwLen);
	if(!*ppBuf)
		return FALSE;

	if(RegQueryValueEx(hKey, pszSubKey, NULL, NULL, (LPBYTE)*ppBuf, &dwLen) != ERROR_SUCCESS)
		return FALSE;

	return TRUE;
}

BOOL LoadShortOption(HKEY hKey, const char *pszSubKey, unsigned short *pShort)
{
	DWORD dwLen, dwType, dwBuf;

	dwLen = sizeof(dwBuf);
	if(RegQueryValueEx(hKey, pszSubKey, NULL, &dwType, (LPBYTE)&dwBuf, &dwLen) != ERROR_SUCCESS)
		return FALSE;

	if(dwType != REG_DWORD || dwLen != sizeof(dwBuf) || dwBuf >> 16)
		return FALSE;

	*pShort = (unsigned short)dwBuf;
	return TRUE;
}

BOOL LoadLongOption(HKEY hKey, const char *pszSubKey, unsigned long *pLong)
{
	DWORD dwLen, dwType;

	dwLen = sizeof(*pLong);
	if(RegQueryValueEx(hKey, pszSubKey, NULL, &dwType, (LPBYTE)pLong, &dwLen) != ERROR_SUCCESS)
		return FALSE;

	if(dwType != REG_DWORD || dwLen != sizeof(*pLong))
		return FALSE;

	return TRUE;
}

BOOL LoadBoolOption(HKEY hKey, const char *pszSubKey, BOOL *pBool)
{
	DWORD dwLen, dwType, dwBuf;

	dwLen = sizeof(dwBuf);
	if(RegQueryValueEx(hKey, pszSubKey, NULL, &dwType, (LPBYTE)&dwBuf, &dwLen) != ERROR_SUCCESS)
		return FALSE;

	if(dwType != REG_DWORD)
		return FALSE;

	if(dwBuf >> 1)
		return FALSE;

	*pBool = (BOOL)dwBuf;
	return TRUE;
}

/* sets the status and all related parameters */
void SetStatus(unsigned int nStatus, char *pStatus, BOOL bReFlash)
{
	/* restore icon handle */
	if(g_hOldIcon)
	{
		g_NID.hIcon = g_hOldIcon;
		/* clear g_hOldIcon so the on/off (flashing) effect may be reproduced */
		g_hOldIcon = NULL;
	}

	/* decide what do to about flashing */
	if((STATUSFLASH(g_nStatus) || bReFlash) && !STATUSFLASH(nStatus))
	{
#if 0
		/* restore icon handle */
		if(g_hOldIcon)
			g_NID.hIcon = g_hOldIcon;

		/* clear g_hOldIcon so the on/off (flashing) effect may be reproduced */
		g_hOldIcon = NULL;
#endif
		/* stop flashing */
		KillTimer(g_hWnd, TIMER_FLASH);
	}
	if((!STATUSFLASH(g_nStatus) || bReFlash) && STATUSFLASH(nStatus))
	{
		/* start flashing */
		SetTimer(g_hWnd, TIMER_FLASH, g_oOptions.uFlash, NULL);
		PostMessage(g_hWnd, WM_TIMER, TIMER_FLASH, 0);
	}

	/* finally, set it */
	g_nStatus = nStatus;

	DestroyIcon(g_NID.hIcon);
	g_NID.hIcon = LoadIcon(g_hInst, MAKEINTRESOURCE(STATUSICON(nStatus)));
	if(pStatus)
		strncpy(g_NID.szTip, pStatus, sizeof(g_NID.szTip));
	else
		g_NID.uFlags &= ~NIF_TIP;
	
	Shell_NotifyIcon(NIM_MODIFY, &g_NID);

	g_NID.uFlags |= NIF_TIP;
}
